from enum import Enum

import numpy as np

from commons.model.exceptions.exceptions import InvalidModelTypeException
from commons.model.linear_regression import LinearRegression


class ModelType(Enum):
    LINEAR_REGRESSION = 1

    @classmethod
    def validate(cls, model_type):
        if not any(model_type == item.name for item in cls):
            raise InvalidModelTypeException(model_type)


class ModelFactory:

    @classmethod
    def get_model(cls, model_type):
        if ModelType[model_type] == ModelType.LINEAR_REGRESSION:
            return LinearRegression
        else:
            raise InvalidModelTypeException(model_type)

    @classmethod
    def load_model(cls, model_type, model_data):
        if ModelType[model_type] == ModelType.LINEAR_REGRESSION:
            return LinearRegression(X=np.asarray(model_data['x']), y=np.asarray(model_data['y']), model_type=model_type)
        else:
            raise InvalidModelTypeException(model_type)



from commons.model.delta_model import DeltaModel
from commons.model.prediction import Prediction
from commons.operations_utils.functions import mean_square_error


class LinearRegression(DeltaModel):
    """Runs linear regression with local data or by gradient steps,
    where gradient can be passed in.

    """
    def __init__(self, X=None, y=None, requirements=None, weights=None):
        self.type = "LINEAR_REGRESSION"
        super().__init__(X, y, self.type, requirements, weights)

    def fit(self, n_iter, eta=0.01):
        """Linear regression for n_iter"""
        for _ in range(n_iter):
            gradient = self.compute_gradient()
            self.gradient_step(gradient, eta)

    def gradient_step(self, gradient, eta=0.01):
        """Update the model with the given gradient"""
        self.weights = self.weights - (eta * gradient)

    def compute_gradient(self):
        """Compute the gradient of the current model using the training set
        """
        prediction = self.predict(self.X)
        delta = prediction.values - self.y
        return delta.dot(self.X) / len(self.X)

    def predict(self, X, y_test=None):
        """Score test data"""
        values = X.dot(self.weights)
        mse = mean_square_error(values, y_test) if y_test is not None else None
        return Prediction(values=values, mse=mse)



import asyncio
import concurrent.futures
import logging

MAX_WORKERS = 5


class AsyncThreadPoolExecutor:

    def __init__(self):
        self.executor = concurrent.futures.ThreadPoolExecutor(max_workers=MAX_WORKERS)

    async def _main(self, executable, args):
        loop = asyncio.get_event_loop()
        futures = [
            loop.run_in_executor(self.executor, executable, arg)
            for arg in args
            ]
        return await asyncio.gather(*futures)

    def run(self, executable, args):
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        results = loop.run_until_complete(self._main(executable, args))
        loop.close()
        return results



from commons.encryption.config import ENCRYPTION_TYPE


class EncryptionService:

    def __init__(self, homomorphic_encryption=ENCRYPTION_TYPE, is_active=False):
        """

        :param homomorphic_encryption:
        """
        self.homomorphic_encryption = homomorphic_encryption()
        self.is_active = is_active
        self.public_key, self.private_key = None, None

    def generate_key_pair(self, key_length):
        """

        :param key_length:
        :return:
        """
        self.public_key, self.private_key = self.homomorphic_encryption.generate_key_pair(key_length=key_length)
        return self.public_key, self.private_key

    def get_public_key(self):
        """

        :return:
        """
        return self.public_key.n

    def get_private_key(self):
        """

        :return:
        """
        return self.private_key

    def set_public_key(self, public_key):
        """

        :param public_key:
        :return:
        """
        self.public_key = self.homomorphic_encryption.get_deserialized_public_key(public_key)

    def encrypt_collection(self, collection, public_key=None):
        """

        :param collection:
        :param public_key:
        :return:
        """
        pk = public_key if public_key else self.public_key
        return self.homomorphic_encryption.encrypt_collection(pk, collection)

    def decrypt_collection(self, collection, private_key=None):
        """

        :param collection:
        :param private_key:
        :return:
        """
        pk = private_key if private_key else self.private_key
        return self.homomorphic_encryption.decrypt_collection(pk, collection)

    def decrypt_and_deserizalize_collection(self, private_key, collection):
        return [self.homomorphic_encryption.decrypt_value(private_key, n) for n in
                self.get_deserialized_collection(collection)]

    def get_serialized_encrypted_collection(self, collection):
        """

        :param collection:
        :return:
        """

        return [self.get_serialized_encrypted_value(value) for value in self.encrypt_collection(collection)]

    def get_serialized_collection(self, collection):
        """

        :param collection:
        :return:
        """
        return [self.get_serialized_encrypted_value(value) for value in collection]

    def get_deserialized_collection(self, collection):
        """

        :param collection:
        :return:
        """
        return [self.get_deserialized_encrypted_value(value) for value in collection]

    def get_serialized_encrypted_value(self, value):
        """

        :param value:
        :return:
        """
        return self.homomorphic_encryption.get_serialized_encrypted_number(value)

    def get_deserialized_encrypted_value(self, value):
        """

        :param value:
        :return:
        """
        return self.homomorphic_encryption.get_encrypted_number(self.public_key, value)

    def get_deserialized_desencrypted_value(self, value):
        return self.homomorphic_encryption.decrypt_value(
            self.private_key,
            self.homomorphic_encryption.get_encrypted_number(self.public_key, value)
        )



class InvalidModelTypeException(Exception):
    def __init__(self, model_type, status_code=400):
        message = "Invalid model type {}".format(model_type)
        # Call the base class constructor with the parameters it needs
        super().__init__(message)

        # Now for your custom code...
        self.status_code = status_code


class ModelErrorException(Exception):
    def __init__(self, status_code=500):
        message = "Error creating model"
        # Call the base class constructor with the parameters it needs
        super().__init__(message)

        # Now for your custom code...
        self.status_code = status_code
